from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from typing import Final

from .diff_proposal import DiffProposal


class ProposalEngine(ABC):
    def __iter__(self):
        return self

    @abstractmethod
    def __next__(self) -> DiffProposal:
        pass


class ProposalList(ProposalEngine):
    def __init__(self, proposals: list[DiffProposal]):
        self.proposals = proposals
        self.index = 0

    def __next__(self) -> DiffProposal:
        if self.index < len(self.proposals):
            prop = self.proposals[self.index]
            self.index += 1
            return prop
        else:
            raise StopIteration


class ProposalRepeat(ProposalEngine):
    def __init__(self, proposal: DiffProposal, times: int = 1):
        self.proposal: Final[DiffProposal] = proposal
        self.index: int = 0
        self.times: Final[int] = times

    def __next__(self) -> DiffProposal:
        if self.index < self.times:
            self.index += 1
            return self.proposal
        else:
            raise StopIteration


class ProposalEngineRepeat(ProposalEngine):
    def __init__(self, proposal_engine: ProposalEngine, times: int = 1):
        self.proposal_engine: Final[ProposalEngine] = proposal_engine
        self.index: int = 0
        self.times: Final[int] = times
        self.proposal_engine_current = copy.deepcopy(self.proposal_engine)

    def __next__(self) -> DiffProposal:
        try:
            return next(self.proposal_engine_current)
        except StopIteration:
            if self.index < self.times:
                self.index += 1
                self.proposal_engine_current = copy.deepcopy(
                    self.proposal_engine)
                return next(self.proposal_engine_current)
            else:
                raise StopIteration



# Copyright: Hallux team, 2023

from __future__ import annotations

import os
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory


@contextmanager
def set_directory(path: str | Path | TemporaryDirectory):
    """Sets the cwd within the context

    Args:
        path (Path): The path to the cwd

    Yields:
        None
    """
    origin = Path().absolute()
    path_str = path.name if isinstance(path, TemporaryDirectory) else str(path)
    try:
        os.chdir(path_str)
        yield
    finally:
        os.chdir(origin)


def find_arg(argv: list[str], name: str) -> int:
    """
    Finds argument index and list of following arguments
    """
    for i in range(len(argv)):
        arg = argv[i]
        if arg == name or arg.startswith(name):
            return i
    return -1



# Copyright: Hallux team, 2023

from __future__ import annotations

from pathlib import Path

from ...issues.issue import IssueDescriptor
from ...proposals.proposal_engine import ProposalEngine, ProposalList
from ...proposals.simple_proposal import SimpleProposal


class CppIssue(IssueDescriptor):
    def __init__(self, filename: str, issue_line: int = 0, description: str = ""):
        super().__init__(
            language="cpp", tool="compile", filename=filename, issue_line=issue_line, description=description
        )
        self.issue_type = "compilation"

    def list_proposals(self) -> ProposalEngine:
        line_comment: str = f" // line {str(self.issue_line)}"
        return ProposalList(
            [
                SimpleProposal(self, radius_or_range=3,
                               issue_line_comment=line_comment),
                SimpleProposal(self, radius_or_range=4,
                               issue_line_comment=line_comment),
                SimpleProposal(self, radius_or_range=5,
                               issue_line_comment=line_comment),
                SimpleProposal(self, radius_or_range=6,
                               issue_line_comment=line_comment),
            ]
        )

    @staticmethod
    def parseMakeIssues(make_output: str, debug: bool = False) -> list[CppIssue]:
        issues: list[CppIssue] = []
        output_lines: list[str] = make_output.split("\n")
        current_issue: CppIssue | None = None

        for line_num in range(len(output_lines)):
            err_line_list = output_lines[line_num].split(":")
            if (
                len(err_line_list) > 4
                and Path(err_line_list[0]).exists()
                and err_line_list[1].isdecimal()
                and err_line_list[2].isdecimal()
                and (err_line_list[3] == " error" or err_line_list[3] == " warning")
            ):
                if current_issue is not None:
                    current_issue.message_lines = current_issue.message_lines[:-1]
                    issues.append(current_issue)

                current_issue = CppIssue(
                    filename=err_line_list[0],
                    issue_line=int(err_line_list[1]),
                    description=str(":".join(err_line_list[4:]).lstrip(" ")),
                )
                current_issue.message_lines = [output_lines[line_num]]
                current_issue.debug = debug

            elif current_issue is not None:
                if output_lines[line_num].startswith("make"):
                    issues.append(current_issue)
                    current_issue = None
                else:
                    current_issue.message_lines.append(output_lines[line_num])

        if current_issue is not None:
            issues.append(current_issue)

        return issues



# Copyright: Hallux team, 2023

from __future__ import annotations

import os
from pathlib import Path
from typing import Final

import requests

from ...backends.query_backend import QueryBackend
from ...targets.diff import DiffTarget
from ..issue_solver import IssueSolver
from .issue import IssueDescriptor, SonarIssue


class OverrideQueryBackend(QueryBackend):
    """
    With SonarQube we only may solve 1 issue per file.
    We override QueryBackend to track files with successful fixes.
    Later, this file list is used to avoid any other fixes
    """

    def __init__(self, underlying_backend: QueryBackend, already_fixed_files: list[str]):
        super().__init__()
        self.backend: Final[QueryBackend] = underlying_backend
        self.already_fixed_files: Final[list[str]] = already_fixed_files

    def query(self, request: str, issue: IssueDescriptor | None = None, issue_lines: list[str] = list) -> list[str]:
        return self.backend.query(request, issue, issue_lines)

    def report_succesfull_fix(self, issue, proposal) -> None:
        self.already_fixed_files.append(issue.filename)
        self.backend.report_succesfull_fix(issue, proposal)


class Sonar_IssueSolver(IssueSolver):
    SONAR_TOKEN: Final[str] = "SONAR_TOKEN"

    def __init__(
        self,
        config_path: Path,
        run_path: Path,
        command_dir: str = ".",
        success_test: str | None = None,
        url: str | None = None,
        token: str | None = None,
        project: str | None = None,
    ):
        super().__init__(config_path, run_path, command_dir, success_test=success_test)

        if token is None:
            if os.getenv(self.SONAR_TOKEN) is not None:
                token = os.getenv(self.SONAR_TOKEN)

        self.token: Final[str | None] = token
        self.url: Final[str | None] = url
        self.project: Final[str | None] = project
        # With SonarQube we only may solve 1 issue per file.
        self.already_fixed_files: Final[list[str]] = []

    def solve_issues(self, diff_target: DiffTarget, query_backend: QueryBackend):
        if not self.token or not self.url or not self.project:
            print("Process SonarQube: not configured")
        else:
            print("Process SonarQube:")
            new_query_backend = OverrideQueryBackend(
                query_backend, self.already_fixed_files)
            super().solve_issues(diff_target, new_query_backend)

    def list_issues(self) -> list[IssueDescriptor]:
        issues: list[IssueDescriptor] = []
        try:
            request: str = (
                self.url
                + "/api/issues/search?resolved=false&severities=MINOR,MAJOR,CRITICAL&statuses=OPEN,CONFIRMED,REOPENED"
            )
            if self.project is not None:
                request += "&componentKeys=" + self.project
            x = requests.get(url=request, headers={
                             "Authorization": "Bearer " + self.token})
        except Exception:
            return []

        if x.status_code == 200:
            issues.extend(SonarIssue.parseIssues(
                x.text, self.already_fixed_files))

        return issues



# Copyright: Hallux team, 2023

from __future__ import annotations

import logging
import subprocess
from pathlib import Path
from typing import Final

from ...auxilary import set_directory
from ...backends.query_backend import QueryBackend
from ...issues.issue import IssueDescriptor
from ...targets.diff import DiffTarget
from ...tools.cpp.issue import CppIssue
from ...tools.issue_solver import IssueSolver


class MakeTargetSolver(IssueSolver):
    def __init__(self, run_path: Path, make_target: str, verbose: bool = False, config_path: Path = Path()):
        super().__init__(config_path=config_path, run_path=run_path)
        # self.makefile_dir: Final[Path] = makefile_dir
        self.make_target: Final[str] = make_target

    def solve_issues(self, diff_target: DiffTarget, query_backend: QueryBackend):
        logging.info(f"{self.run_path}/Makefile : '{self.make_target}'")
        with set_directory(self.run_path):
            super().solve_issues(diff_target, query_backend)

    def list_issues(self) -> list[IssueDescriptor]:
        issues: list[IssueDescriptor] = []

        with set_directory(self.run_path):
            try:
                subprocess.check_output(
                    ["make", self.make_target], stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                make_output: str = e.output.decode("utf-8")
                issues.extend(CppIssue.parseMakeIssues(make_output))

        return issues



# Copyright: Hallux team, 2023

from __future__ import annotations

import os
import subprocess

from github import Github, GithubException, PullRequest, Repository

from ..proposals.diff_proposal import DiffProposal
from .filesystem import FilesystemTarget


# Saves Issue Fixes as Github proposals
class GithubProposalTraget(FilesystemTarget):
    def __init__(self, pr_url: str):
        FilesystemTarget.__init__(self)

        (base_url, repo_name, PR_ID) = GithubProposalTraget.parse_pr_url(pr_url)

        if base_url is None:
            raise SystemError(f"Cannot parse github PR URL: {pr_url}")

        if "GITHUB_TOKEN" not in os.environ.keys():
            raise SystemError("GITHUB_TOKEN is not provided")

        self.github = Github(os.environ["GITHUB_TOKEN"], base_url=base_url)
        self.repo: Repository = self.github.get_repo(repo_name)
        self.pull_request: PullRequest = self.repo.get_pull(PR_ID)

        if self.pull_request.closed_at is not None:  # self.pull_request.is_merged :
            raise SystemError(
                f"Pull Request {PR_ID} is either closed or merged already")

        latest_github_sha: str = self.pull_request.head.sha
        git_log = subprocess.check_output(
            ["git", "log", "--pretty=oneline"]).decode("utf8")
        local_git_sha = git_log.split("\n")[0].split(" ")[0]

        if local_git_sha != latest_github_sha:
            raise SystemError(
                f"Local git commit `{local_git_sha}` and head from pull-request `{latest_github_sha}` do not coincide!"
            )

    @staticmethod
    def parse_pr_url(pr_url: str) -> tuple[str, str, int] | tuple[None, None, None]:
        """
        Tries to parse Pull-Request URL to obtain base_url, repo_name and PR_ID
        :param pr_url: shall look like this: https://BUSINESS.github.com/YOUR_NAME/REPO_NAME/pull/ID
        :return: (base_url, base_url = "https://api." + url_items[0], PR_ID) OR None
        """
        if pr_url.startswith("https://"):
            pr_url = pr_url[len("https://"):]
            url_items = pr_url.split("/")
            if len(url_items) == 5 and url_items[3] == "pull":
                if url_items[0] == "github.com":
                    base_url = "https://api." + url_items[0]
                else:
                    base_url = "https://api." + url_items[0] + "/api/v3"
                repo_name = url_items[1] + "/" + url_items[2]
                pr_id = int(url_items[4])
                return base_url, repo_name, pr_id
        return None, None, None

    def apply_diff(self, diff: DiffProposal) -> bool:
        for file in self.pull_request.get_files():
            if diff.filename == file.filename:
                FilesystemTarget.apply_diff(self, diff)
                return True
        return False

    def revert_diff(self):
        FilesystemTarget.revert_diff(self)

    def commit_diff(self) -> bool:
        commit = self.repo.get_commit(self.pull_request.head.sha)
        body = self.existing_proposal.description + "\n```suggestion\n"
        for line in self.existing_proposal.proposed_lines:
            body = body + line + "\n"
        body = body + "\n```"

        success: bool = True
        try:
            self.pull_request.create_review_comment(
                body=body,
                commit=commit,
                side="RIGHT",
                path=str(self.existing_proposal.filename),
                line=self.existing_proposal.end_line,
                start_line=self.existing_proposal.start_line,
            )
        except GithubException:
            success = False

        # clean local code
        FilesystemTarget.revert_diff(self)
        return success

    def requires_refresh(self) -> bool:
        return False



# Copyright: Hallux team, 2023

from __future__ import annotations

import logging
import os
import subprocess
from pathlib import Path

from ..proposals.diff_proposal import DiffProposal
from .filesystem import FilesystemTarget


class GitCommitTarget(FilesystemTarget):
    """
    Saves fixes into local git repo as individual commits
    """

    def __init__(self):
        # ToDo: assert we're in GIT repo, crash if not
        FilesystemTarget.__init__(self)
        git_status_output = subprocess.check_output(
            ["git", "status", "--porcelain"]).decode("utf8")
        if len(git_status_output) > 0:
            raise SystemError(
                "for GIT TARGET you must be in the GIT REPO with no local uncommitted changes!")

    def apply_diff(self, diff: DiffProposal) -> bool:
        return FilesystemTarget.apply_diff(self, diff)

    def revert_diff(self) -> None:
        FilesystemTarget.revert_diff(self)

    def commit_diff(self) -> bool:
        curr_dir: str = os.getcwd()
        git_dir: str = str(Path(self.existing_proposal.filename).parent)
        os.chdir(git_dir)
        success: bool = True
        try:
            logging.debug(
                f"git add {os.path.relpath(self.existing_proposal.filename, start=git_dir)}")
            output = subprocess.check_output(
                ["git", "add", os.path.relpath(
                    self.existing_proposal.filename, start=git_dir)]
            )
            git_message = "HALLUX: " + \
                self.existing_proposal.description.replace('"', "")

            logging.debug(output.decode("utf8"))
            logging.debug(f"git commit -m {git_message}")

            output = subprocess.check_output(
                ["git", "commit", "-m", f"{git_message}"])

            logging.debug(output.decode("utf8"))
            FilesystemTarget.commit_diff(self)
        except subprocess.CalledProcessError as e:
            logging.debug("ERROR:")
            logging.debug(e.output.decode("utf8"))
            FilesystemTarget.revert_diff(self)
            success = False
        finally:
            os.chdir(curr_dir)

        return success

    def requires_refresh(self) -> bool:
        return True



# Copyright: Hallux team, 2023

from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Final

import openai
from openai.api_resources import ChatCompletion

from ..backends.query_backend import QueryBackend
from ..issues.issue import IssueDescriptor


class OpenAiChatGPT(QueryBackend):
    def __init__(
        self,
        model: str = "",
        max_tokens: int = 4097,
        type="openai",
        base_path: Path = Path(),
        previous_backend: QueryBackend | None = None,
    ):
        assert type == "openai"
        super().__init__(base_path, previous_backend)
        self.valid = True
        if model is None or len(model) < 2:
            logging.warning(f"Wrong model name for OpenAI Backend: {model}")
            self.valid = False

        if os.getenv("OPENAI_API_KEY") is None:
            logging.warning(
                "Environment variable OPENAI_API_KEY is required for OpenAI Backend")
            self.valid = False

        self.model: Final[str] = model
        self.max_tokens: Final[int] = max_tokens

        openai.api_key = os.getenv("OPENAI_API_KEY")

    def query(self, request: str, issue: IssueDescriptor | None = None, issue_lines: list[str] = list) -> list[str]:
        if not self.valid:
            return []

        logging.debug("[OpenAI REQUEST]:")
        logging.debug(request)
        result = ChatCompletion.create(
            messages=[{"role": "user", "content": request}], model=self.model)
        answers = []
        if len(result["choices"]) > 0:
            for variant in result["choices"]:
                answers.append(variant["message"]["content"])

        logging.debug("[OpenAI ANSWERS]:")
        for ans in answers:
            for line in ans.split("\n"):
                logging.debug(line)

        return answers



# Copyright: Hallux team, 2023

# We have several options where (and how) to send found fixes (aka FileDiff).
# 1. FilesystemTarget (most primitive one) - it simply to re-write local files
# 2. GitCommitTarget (little more complicated) - rewrite local files, but also make git commit on every change
# 3. GithubProposalTarget - submit all changes into Github Web Interface as proposals/suggestions
#    This last one requires some extra configuring to be done
# All those approaches are hidden after DiffTarget interface written in this file

from __future__ import annotations

from abc import ABC, abstractmethod

from ..proposals.diff_proposal import DiffProposal


# Interface for DiffTarget implementations
# It has some kind of state:
# * DiffProposal needs to be firstly applied (temporally),
# * then checked/tested by some other means,
# * if check was successful DiffProposal shall be finally committed
# * otherwise reverted
class DiffTarget(ABC):
    @abstractmethod
    def apply_diff(self, diff: DiffProposal) -> bool:
        pass

    @abstractmethod
    def revert_diff(self) -> None:
        pass

    @abstractmethod
    def commit_diff(self) -> bool:
        pass

    @abstractmethod
    def requires_refresh(self) -> bool:
        """
        :return: true if issues need to be refreshed, after successful solve.
                 Could be due-to line-numbers change, or other reasons.
        """
        pass



# Copyright: Hallux team, 2023

from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Any

import requests

from ..backends.query_backend import QueryBackend
from ..issues.issue import IssueDescriptor


class RestBackend(QueryBackend):
    PROMPT_STRING = "$PROMPT"

    def __init__(
        self,
        url: str,
        request_body: str | dict = PROMPT_STRING,
        response_body: str = "$RESPONSE",
        token: str | None = None,
        type="rest",
        base_path: Path = Path(),
        previous_backend: QueryBackend | None = None,
    ):
        super().__init__(base_path, previous_backend)
        assert type == "rest"
        self.url = url
        self.token = token
        self.method = "POST"
        self.headers = {}
        self.request_body = request_body
        self.response_body = response_body

    def _is_string(self, obj):
        if isinstance(obj, str):
            # If obj is a string, it's a valid object
            return True
        return False

    def _is_object(self, obj):
        if isinstance(obj, (list, dict, tuple)):  # Add other valid object types as needed
            # If obj is a list, dict, or tuple, it's a valid object
            return True
        return False

    def _make_request(self, request: str):
        try:
            response = requests.post(self.url, json={"message": request})

            # Successful response
            if response.status_code == 200:
                if response.headers["Content-Type"] == "application/json":
                    return response.json()
                else:
                    return response.text

            else:
                logging.warning(f"Error status code: {response.status_code}")

        except requests.ConnectionError:
            logging.warning(f"Host {self.url} is not reachable.")

        return None

    def _get_by_path(self, response: str | None | dict):
        keys = self.response_body.split(".")[1:]

        current = response

        try:
            for key in keys:
                if isinstance(current, dict):
                    current = current[key]
                elif isinstance(current, list):
                    if key.isdigit():
                        key = int(key)
                        current = current[key]
                    else:
                        return []
                else:
                    return []
        except (KeyError, IndexError, ValueError):
            return []

        return [current]

    """
    Takes the response object and response_body configuration in the following format:
    "$RESPONSE.answer.0.value" and returns the value of the response object at that path.
    """

    def _parse_response(self, response: str | None | dict) -> list[str]:
        if response is None:
            return []

        if self.response_body == "$RESPONSE" and self._is_string(response):
            return [response]

        elif self.response_body.startswith("$RESPONSE."):
            return self._get_by_path(response)

        else:
            return []

    def query(self, request: str, issue: IssueDescriptor | None = None, issue_lines: list[str] = list) -> list[str]:
        if self._is_object(self.request_body):
            json_data = json.dumps(self.request_body)
            parsed_request = json_data.replace(self.PROMPT_STRING, request)
            self.headers.update({"Content-Type": "application/json"})

        elif self._is_string(self.request_body):
            parsed_request = self.request_body.replace(
                self.PROMPT_STRING, request)
            self.headers.update({"Content-Type": "text/plain"})

        else:
            logging.warning("Invalid request body")
            return []

        logging.debug("RestBackend REQUEST")
        logging.debug(parsed_request)

        response = self._make_request(parsed_request)
        parsed_response = self._parse_response(response)

        logging.debug("RestBackend ANSWERS")
        logging.debug(parsed_response)

        return parsed_response



# Copyright: Hallux team, 2023

from __future__ import annotations

from pathlib import Path

from ..auxilary import find_arg
from ..tools.cpp.cpp import Cpp_IssueSolver
from ..tools.issue_solver import IssueSolver
from ..tools.mypy.solver import Mypy_IssueSolver
from ..tools.ruff.solver import Ruff_IssueSolver
from ..tools.sonarqube.solver import Sonar_IssueSolver


class ProcessorFactory:
    @staticmethod
    def init_solvers(
        argv: list[str],
        config: dict | None,
        groups: dict[str, list[str]] | None,
        config_path: Path,
        run_path: Path,
        command_dir: str = ".",
    ) -> list[IssueSolver]:
        config = config if config is not None else {}
        mapping: dict = {
            "ruff": Ruff_IssueSolver,
            "mypy": Mypy_IssueSolver,
            "sonar": Sonar_IssueSolver,
            "cpp": Cpp_IssueSolver,
        }

        groups = (
            groups
            if groups is not None
            else {
                "all": list(mapping),
                "python": ["ruff", "mypy"],
            }
        )

        requested_names: dict[str] = {}

        # first check for any particular group(s)
        for group_name, group_list in groups.items():
            if find_arg(argv, "--" + group_name) > 0:
                for name in group_list:
                    requested_names[name] = True

        # if not groups asked, check for individual processors
        if len(requested_names) == 0:
            for name in mapping.keys():
                if find_arg(argv, "--" + name) > 0:
                    requested_names[name] = True

        # if nothing particular asked - just use all of them
        if len(requested_names) == 0:
            requested_names = mapping

        solvers: list[IssueSolver] = []
        for name in requested_names:
            classname = mapping[name]
            config_params = config.get(name, {})
            solver = classname(**config_params, config_path=config_path,
                               run_path=run_path, command_dir=command_dir)
            solvers.append(solver)

        return solvers



# Copyright: Hallux team, 2023

from __future__ import annotations

from pathlib import Path

from ..auxilary import find_arg
from ..backends.dummy_backend import DummyBackend
from ..backends.openai_backend import OpenAiChatGPT, QueryBackend
from ..backends.rest_backend import RestBackend


class BackendFactory:
    @staticmethod
    def init_backend(argv: list[str], backends_list: list[dict] | None, config_path: Path) -> QueryBackend:
        default_list = [
            {"cache": {"type": "dummy", "filename": "dummy.json"}},
            {"rest": {"type": "rest", "url": "http://localhost:8000/generate"}},
            {"gpt3": {"type": "openai", "model": "gpt-3.5-turbo", "max_tokens": 4096}},
            # Gonna be enabled soon {"gpt4": {"type": "openai", "model": "gpt-4", "max_tokens": 8192}},
        ]

        backends_list: list[dict] = backends_list if backends_list is not None else default_list
        backend = None
        if not isinstance(backends_list, list) or len(backends_list) == 0:
            raise SystemError(
                f"BACKENDS config setting must contain non-empty list. Error in: {backends_list}")

        for name_dict in backends_list:
            name, settings = BackendFactory._validate_settings(name_dict)
            backend = BackendFactory._create_backend(
                settings, config_path, backend)
            if find_arg(argv, "--" + name) > 0:
                return backend  # stop early if required by CLI
        return backend

    @staticmethod
    def _validate_settings(name_dict: dict) -> tuple[str, dict[str, str]]:
        if not isinstance(name_dict, dict) or len(name_dict.items()) != 1:
            raise SystemError(
                f"Each BACKEND must be just one name-settings pair. Error in: {name_dict}")

        name = str(list(name_dict)[0])
        settings = name_dict[name]

        if not isinstance(settings, dict) or "type" not in settings.keys():
            raise SystemError(
                f"Each BACKEND setting must have at least 'type' setting. Error in: {settings}")
        return name, settings

    @staticmethod
    def _create_backend(settings: dict, config_path: Path, previous_backend: QueryBackend) -> QueryBackend:
        type_to_class = {"dummy": DummyBackend,
                         "openai": OpenAiChatGPT, "rest": RestBackend}
        backend_type = settings["type"]
        backend_class = type_to_class.get(backend_type)
        if backend_class is None:
            raise SystemError(f"Unknown BACKEND type: {backend_type}")
        return backend_class(**settings, base_path=config_path, previous_backend=previous_backend)
